SorhaQuestLog = LibStub("AceAddon-3.0"):NewAddon("SorhaQuestLog", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SorhaQuestLog")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local db
local dbChar

BINDING_HEADER_SORHAQUESTLOG = "Sorha Quest Log";
BINDING_NAME_TOGGLE_SORHAQUESTLOG = L["Toggle Minion"]

--Frames
local fraSQLMainAnchor = nil
local dicButtons = {}
local intNumberOfFrames = 50
local strButtonPrefix = "SorhaQuestLogButton"

--Colour strings
local strHeaderColour = "|cff00aaff"
local strObjectiveColour = "|cffffffff"
local strDoneObjectiveColour = "|cff00ffff"
local strObjectiveDescriptionColour = "|cff888888"

--Just some booleans
local blnInitialized = false
local blnIsUpdating = false
local blnWasAClick = false

local defaults = { --Defaults Table
	profile = {
		Frames = {
			MainWindowX = 0,
			MainWindowY = 0,
			MainWindowPoint = "CENTER",
			MainWindowRelativePoint = "CENTER",
			MainWindowScale = 1,
			MainWindowLocked = false,			
			QuestFontSize = 11,
			HeaderFontSize = 11,
			MinionWidth = 220,
			HeaderColourRed = 0,
			HeaderColourGreen = 0.6,
			HeaderColourBlue = 1,
			HeaderColourAlpha = 1,
			ObjectiveColourRed = 1,
			ObjectiveColourGreen = 1,
			ObjectiveColourBlue = 1,
			ObjectiveColourAlpha = 1,
			ObjectiveDoneColourRed = 0,
			ObjectiveDoneColourGreen = 1,
			ObjectiveDoneColourBlue = 1,
			ObjectiveDoneColourAlpha = 1,		
			ObjectiveDescColourRed = 0.5,
			ObjectiveDescColourGreen = 0.5,
			ObjectiveDescColourBlue = 0.5,
			ObjectiveDescColourAlpha = 1,	
			ShowItemButtons = false,
			HideCompletedObjectives = false,
			ShowNumberOfQuests = false,
			MoveItemsAndTooltipsRight = false,
			ShowDescWhenNoObjectives = false,
			HideSorhaQuestLog = false,
			CollapseToLeft = false,
			ShowAllQuests = false,
			AllowHiddenQuests = false,
		},
	},
	char = {
		Quests = {
			HiddenQuests = {},
		},
	},
}

local function getOptions() --Creates Options table.
	if not SorhaQuestLog.options then
		SorhaQuestLog.options = {
			type = "group",
			name = "SorhaQuestLog",
			handler = SorhaQuestLog,
			args = {
				General = {
					name = L["General Options"],
					type = "group",
					order = 1,
					args = {
						MainLock = {
							name = L["Lock Frame"],
							desc = L["Unlocks/Locks the Main Frame"],
							type = "toggle",
							get = function() return db.Frames.MainWindowLocked end,
							set = function()
										db.Frames.MainWindowLocked = not db.Frames.MainWindowLocked
										SorhaQuestLog:doAnchorBackgroundUpdate()
									end,
							order = 1,
						},
						MainHide = {
							name = L["Hide Minion"],
							desc = L["Hides/Shows the Main Frame"],
							type = "toggle",
							get = function() return db.Frames.HideSorhaQuestLog end,
							set = function()
										SorhaQuestLog:ToggleSorhaQuestLog()
									end,
							order = 2,
						},
						Reset = {
							order = 3,
							type = "execute",
							name = L["Reset Main Frame"],
							desc = L["Resets Main Frame position"],
							func = function()
								db.Frames.MainWindowPoint = "CENTER"
								db.Frames.MainWindowRelativePoint =  "CENTER"
								db.Frames.MainWindowX = 0
								db.Frames.MainWindowY = 0
							
								fraSQLMainAnchor:StopMovingOrSizing()		
								fraSQLMainAnchor:ClearAllPoints()
								fraSQLMainAnchor:SetPoint(db.Frames.MainWindowPoint, UIParent, db.Frames.MainWindowRelativePoint, db.Frames.MainWindowX, db.Frames.MainWindowY)
							end,
						},
						MainWidth = {
							order = 4,
							name = L["Width"],
							desc = L["Controls the width of the main window."],
							type = "range",
							min = 150, max = 300, step = 1,
							isPercent = false,
							get = function() return db.Frames.MinionWidth end,
							set = function(info, value)
								db.Frames.MinionWidth = value
								fraSQLMainAnchor:SetWidth(db.Frames.MinionWidth)
								SorhaQuestLog:UpdateMinion()
							end,
						},	
						ShowNumberQuestsToggle = {
							name = L["Show number of quests"],
							desc = L["Shows/Hides the number of quests"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.ShowNumberOfQuests end,
							set = function()
								db.Frames.ShowNumberOfQuests = not db.Frames.ShowNumberOfQuests
								SorhaQuestLog:UpdateMinion()
							end,
							order = 5,
						},
						CollapseToLeftToggle = {
							name = L["Autoshrink to left"],
							desc = L["Shrinks the width down when the length of current quests is less then the max width\nNote: Doesn't work well with quests that wordwrap"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.CollapseToLeft end,
							set = function()
								db.Frames.CollapseToLeft = not db.Frames.CollapseToLeft
								SorhaQuestLog:UpdateMinion()
							end,
							order = 6,
						},						
					},
				},
				QuestOptions = {
					name = L["Quest Options"],
					type = "group",
					order = 1,
					args = {
						HeaderFontSize = {
							order = 1,
							name = L["Header Font Size"],
							desc = L["Controls the font size of the header label."],
							type = "range",
							min = 10, max = 20, step = 1,
							isPercent = false,
							get = function() return db.Frames.HeaderFontSize end,
							set = function(info, value)
								db.Frames.HeaderFontSize = value
								SorhaQuestLog:UpdateMinion()
							end,
						},	
						QuestFontSize = {
							order = 2,
							name = L["Quest Font Size"],
							desc = L["Controls the font size of the quest info."],
							type = "range",
							min = 10, max = 20, step = 1,
							isPercent = false,
							get = function() return db.Frames.QuestFontSize end,
							set = function(info, value)
								db.Frames.QuestFontSize = value
								SorhaQuestLog:UpdateMinion()
							end,
						},	
						HideCompletedObjectivesToggle = {
							name = L["Hide completed objectives"],
							desc = L["Shows/Hides completed objectives"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.HideCompletedObjectives end,
							set = function()
								db.Frames.HideCompletedObjectives = not db.Frames.HideCompletedObjectives
								SorhaQuestLog:UpdateMinion()
							end,
							order = 3,
						},
						ShowItemButtonsToggle = {
							name = L["Show quest item buttons"],
							desc = L["Shows/Hides the quest item buttons"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.ShowItemButtons end,
							set = function()
								db.Frames.ShowItemButtons = not db.Frames.ShowItemButtons
								SorhaQuestLog:UpdateMinion()
							end,
							order = 4,
						},
						ItemsAndTooltipsRightToggle = {
							name = L["Display items and tooltips on right"],
							desc = L["Moves items and tooltips to the right"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.MoveItemsAndTooltipsRight end,
							set = function()
								db.Frames.MoveItemsAndTooltipsRight = not db.Frames.MoveItemsAndTooltipsRight
								SorhaQuestLog:UpdateMinion()
							end,
							order = 5,
						},
						ShowDescWhenNoObjectivesToggle = {
							name = L["Display quest description if not objectives"],
							desc = L["Displays a quests description if there are no objectives available"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.ShowDescWhenNoObjectives end,
							set = function()
								db.Frames.ShowDescWhenNoObjectives = not db.Frames.ShowDescWhenNoObjectives
								SorhaQuestLog:UpdateMinion()
							end,
							order = 6,
						},
						AllowHiddenQuestsToggle = {
							name = L["Allow quests to be hidden"],
							desc = L["Allows quests to be hidden and enables the show/hide button"],
							type = "toggle",
							width = "full",
							get = function() return db.Frames.AllowHiddenQuests end,
							set = function()
								db.Frames.AllowHiddenQuests = not db.Frames.AllowHiddenQuests
								SorhaQuestLog:doHiddenQuestsUpdate()
							end,
							order = 7,
						},						
					},
				},
				ColourOptions = {
					name = L["Colour Options"],
					type = "group",
					order = 2,
					args = {		
						HeaderColour = {
							name = L["Zone Header Colour"],
							desc = L["Sets the color for the header of each zone"],
							type = "color",
							width = "full",
							hasAlpha = true,
							get = function() return db.Frames.HeaderColourRed, db.Frames.HeaderColourGreen, db.Frames.HeaderColourBlue, db.Frames.HeaderColourAlpha end,
							set = function(_,r,g,b,a)
									db.Frames.HeaderColourRed = r
									db.Frames.HeaderColourGreen = g
									db.Frames.HeaderColourBlue = b
									db.Frames.HeaderColourAlpha = a
									SorhaQuestLog:doColourChange()
								end,
							order = 1,
						},
						UndoneObjectiveColour = {
							name = L["Incomplete objective colour"],
							desc = L["Sets the color for the incomplete objectives"],
							type = "color",
							width = "full",
							hasAlpha = true,
							get = function() return db.Frames.ObjectiveColourRed, db.Frames.ObjectiveColourGreen, db.Frames.ObjectiveColourBlue, db.Frames.ObjectiveColourAlpha end,
							set = function(_,r,g,b,a)
									db.Frames.ObjectiveColourRed = r
									db.Frames.ObjectiveColourGreen = g
									db.Frames.ObjectiveColourBlue = b
									db.Frames.ObjectiveColourAlpha = a
									SorhaQuestLog:doColourChange()
								end,
							order = 2,
						},
						DoneObjectiveColour = {
							name = L["Complete objective colour"],
							desc = L["Sets the color for the complete objectives"],
							type = "color",
							width = "full",
							hasAlpha = true,
							get = function() return db.Frames.ObjectiveDoneColourRed, db.Frames.ObjectiveDoneColourGreen, db.Frames.ObjectiveDoneColourBlue, db.Frames.ObjectiveDoneColourAlpha end,
							set = function(_,r,g,b,a)
									db.Frames.ObjectiveDoneColourRed = r
									db.Frames.ObjectiveDoneColourGreen = g
									db.Frames.ObjectiveDoneColourBlue = b
									db.Frames.ObjectiveDoneColourAlpha = a
									SorhaQuestLog:doColourChange()
								end,
							order = 3,
						},
						NoObjectivesColour = {
							name = L["No objectives description colour"],
							desc = L["Sets the color for the description displayed when there is no quest objectives"],
							type = "color",
							width = "full",
							hasAlpha = true,
							get = function() return db.Frames.ObjectiveDescColourRed, db.Frames.ObjectiveDescColourGreen, db.Frames.ObjectiveDescColourBlue, db.Frames.ObjectiveDescColourAlpha end,
							set = function(_,r,g,b,a)
									db.Frames.ObjectiveDescColourRed = r
									db.Frames.ObjectiveDescColourGreen = g
									db.Frames.ObjectiveDescColourBlue = b
									db.Frames.ObjectiveDescColourAlpha = a
									SorhaQuestLog:doColourChange()
								end,
							order = 4,
						},
					},
				},
			},
		}
		SorhaQuestLog.options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(SorhaQuestLog.db)
		SorhaQuestLog.options.args.profiles.order = 9
	end
	return SorhaQuestLog.options
	
end

function SorhaQuestLog:OnInitialize() --Called when the addon is loaded
	self.db = LibStub("AceDB-3.0"):New("SorhaQuestLogDB", defaults, true)
	db = self.db.profile
	dbChar = self.db.char
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
	
	self:SetupOptions()
	self:SetupColours()
end

function SorhaQuestLog:OnEnable() --Called when the addon is enabled
	self:CreateLayout()
	self:RegisterEvent("QUEST_LOG_UPDATE")	
	self:RegisterEvent('PLAYER_LEVEL_UP');
	blnInitialized = true
end

function SorhaQuestLog:RefreshConfig()
	db = self.db.profile
	fraSQLMainAnchor:SetWidth(db.Frames.MinionWidth)
	self:doColourChange()
	self:doAnchorBackgroundUpdate()
	self:UpdateMinion()
end

function SorhaQuestLog:CreateLayout()
	fraSQLMainAnchor = self:doCreateFrame("FRAME","SorhaQuestLogMainAnchor",UIParent,db.Frames.MinionWidth,40,1,"BACKGROUND",1,db.Frames.MainWindowPoint, UIParent, db.Frames.MainWindowRelativePoint, db.Frames.MainWindowX, db.Frames.MainWindowY,1)
	fraSQLMainAnchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,	insets = {left = 5, right = 3, top = 3, bottom = 5}})
	fraSQLMainAnchor:SetMovable(true)
	fraSQLMainAnchor:SetClampedToScreen(true)
	fraSQLMainAnchor:RegisterForDrag("LeftButton")
	fraSQLMainAnchor:SetScript("OnDragStart", fraSQLMainAnchor.StartMoving)
	fraSQLMainAnchor:SetScript("OnDragStop",  function(self)
		fraSQLMainAnchor:StopMovingOrSizing()
		strPoint, tempB, strRelativePoint, intPosX, intPosY = fraSQLMainAnchor:GetPoint()
		db.Frames.MainWindowPoint = strPoint
		db.Frames.MainWindowRelativePoint = strRelativePoint
		db.Frames.MainWindowX = intPosX
		db.Frames.MainWindowY = intPosY
	end)
	fraSQLMainAnchor.objFontString = fraSQLMainAnchor:CreateFontString(nil, "OVERLAY");
	fraSQLMainAnchor.objFontString:SetFont("Interface\\AddOns\\SorhaQuestLog\\Fonts\\framd.ttf", 14)
	fraSQLMainAnchor.objFontString:SetJustifyH("LEFT")
	fraSQLMainAnchor.objFontString:SetJustifyV("TOP")
	fraSQLMainAnchor.objFontString:SetText("");
	fraSQLMainAnchor.objFontString:SetShadowColor(0.0, 0.0, 0.0, 1.0)
	fraSQLMainAnchor.objFontString:SetShadowOffset(1, -1)
	
	fraSQLMainAnchor.buttonShowHidden = self:doCreateLooseFrame("BUTTON","SQLShowHiddenButton",fraSQLMainAnchor,16,16,1,"LOW",1,1)
	fraSQLMainAnchor.buttonShowHidden:SetPoint("TOPLEFT", fraSQLMainAnchor, "TOPLEFT", 0, -8);
	fraSQLMainAnchor.buttonShowHidden:SetBackdrop({bgFile="Interface\\AddOns\\SorhaQuestLog\\Textures\\button1.tga", edgeFile="Interface\\Tooltips\\UI-Tooltip-Border", tile=0, tileSize=0, edgeSize=1, insets={left=-0, right=0, top=0, bottom=0}})
	fraSQLMainAnchor.buttonShowHidden:SetBackdropBorderColor(0, 0, 0, 0)
	
	
	fraSQLMainAnchor.buttonShowHiddenHover = self:doCreateLooseFrame("FRAME","SQLShowHiddenButtonHover",fraSQLMainAnchor.buttonShowHidden,fraSQLMainAnchor.buttonShowHidden:GetWidth(),fraSQLMainAnchor.buttonShowHidden:GetHeight(),1,"LOW",1,1)
	fraSQLMainAnchor.buttonShowHiddenHover:SetPoint("TOPLEFT", fraSQLMainAnchor.buttonShowHidden, "TOPLEFT", 0, 0);
	fraSQLMainAnchor.buttonShowHiddenHover:SetBackdrop({bgFile="Interface\\AddOns\\SorhaQuestLog\\Textures\\button1_hover.tga", edgeFile="Interface\\Tooltips\\UI-Tooltip-Border", tile=0, tileSize=0, edgeSize=1, insets={left=-0, right=0, top=0, bottom=0}})
	fraSQLMainAnchor.buttonShowHiddenHover:SetBackdropColor(0, 0, 1, 1)
	fraSQLMainAnchor.buttonShowHiddenHover:SetBackdropBorderColor(0,0,0,0)
	fraSQLMainAnchor.buttonShowHiddenHover:SetAlpha(0)
	
	
	fraSQLMainAnchor.buttonShowHidden:RegisterForClicks("AnyUp")
	fraSQLMainAnchor.buttonShowHidden:SetScript("OnEnter", function() fraSQLMainAnchor.buttonShowHiddenHover:SetAlpha(1) end)
	fraSQLMainAnchor.buttonShowHidden:SetScript("OnLeave", function() fraSQLMainAnchor.buttonShowHiddenHover:SetAlpha(0) end)
	fraSQLMainAnchor.buttonShowHidden:SetScript("OnClick", function() 
		db.Frames.ShowAllQuests = not db.Frames.ShowAllQuests
		if (db.Frames.ShowAllQuests == true) then
			fraSQLMainAnchor.buttonShowHidden:SetBackdropColor(0, 0.3, 1, 1)
		else
			fraSQLMainAnchor.buttonShowHidden:SetBackdropColor(0.5, 0.5, 0.5, 1)
		end
		self:UpdateMinion()
	end)

	
	self:doAnchorBackgroundUpdate()
	self:doHiddenQuestsUpdate()
	self:ShowSorhaQuestLog()
	
	local strButtonName = ""
	local i = 1
	for i=1,intNumberOfFrames,1 do
		strButtonName = strButtonPrefix .. i
		dicButtons[strButtonPrefix .. i] = self:doCreateLooseFrame("BUTTON",strButtonName,fraSQLMainAnchor,10,10,1,"LOW",1,1)
		dicButtons[strButtonPrefix .. i].objFontString = dicButtons[strButtonPrefix .. i]:CreateFontString(nil, "OVERLAY");
		dicButtons[strButtonPrefix .. i].objFontString:SetFont("Interface\\AddOns\\SorhaQuestLog\\Fonts\\framd.ttf", 11)
		dicButtons[strButtonPrefix .. i].objFontString:SetPoint("LEFT", dicButtons[strButtonPrefix .. i], "LEFT", 0, 0);
		dicButtons[strButtonPrefix .. i].objFontString:SetJustifyH("LEFT")
		dicButtons[strButtonPrefix .. i].objFontString:SetJustifyV("TOP")
		dicButtons[strButtonPrefix .. i].objFontString:SetText("");
		dicButtons[strButtonPrefix .. i].objFontString:SetShadowColor(0.0, 0.0, 0.0, 1.0)
		dicButtons[strButtonPrefix .. i].objFontString:SetShadowOffset(1, -1)
		dicButtons[strButtonPrefix .. i]:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,	insets = {left = 5, right = 3, top = 3, bottom = 5}})
		dicButtons[strButtonPrefix .. i]:SetBackdropColor(0, 0, 0, 0)
		dicButtons[strButtonPrefix .. i]:SetBackdropBorderColor(0, 0, 0, 0)
		
		

		strButtonName = strButtonPrefix .. "item" .. i
		dicButtons[strButtonPrefix .. i].fraItemButton = CreateFrame("BUTTON", strButtonName, dicButtons[strButtonPrefix .. i], "WatchFrameItemButtonTemplate");
		dicButtons[strButtonPrefix .. i].fraItemButton:SetScale(0.75)

		dicButtons[strButtonPrefix .. i].isHeader = nil
		dicButtons[strButtonPrefix .. i].isCollapsed = nil
		dicButtons[strButtonPrefix .. i].isUsed = false
		
		dicButtons[strButtonPrefix .. i]:RegisterForClicks("AnyUp")
		dicButtons[strButtonPrefix .. i]:SetScript("OnClick", function(self, button)
			blnWasAClick = true
			if (button == "LeftButton") then
				if (dicButtons[strButtonPrefix .. i].isHeader == 1) then
					if (dicButtons[strButtonPrefix .. i].isCollapsed == 1) then
						ExpandQuestHeader(i)
					else
						CollapseQuestHeader(i)
					end
				else
					if (IsShiftKeyDown()) then
						ChatEdit_InsertLink(GetQuestLink(i))
					else
						ShowUIPanel(QuestLogFrame);
						SelectQuestLogEntry(i);
						QuestLog_SetSelection(i);
						QuestLog_Update();
					end
				end
			else
				local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(i);
				if not(dicButtons[strButtonPrefix .. i].isHeader == 1) then
					if not(tContains(dbChar.Quests.HiddenQuests, questID)) then
						tinsert(dbChar.Quests.HiddenQuests, questID)	
					else
						SorhaQuestLog:RemoveObject(dbChar.Quests.HiddenQuests, questID)
					end
					SorhaQuestLog:UpdateMinion()
				end
			end
		end)
		dicButtons[strButtonPrefix .. i]:SetScript("OnEnter", function(self)
			if (not(dicButtons[strButtonPrefix .. i].isHeader)) then
				SelectQuestLogEntry(i);
				local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(i);
				if (db.Frames.MoveItemsAndTooltipsRight == true) then
					GameTooltip:SetOwner(dicButtons[strButtonPrefix .. i], "ANCHOR_RIGHT", 0, -50);
				else 
					GameTooltip:SetOwner(dicButtons[strButtonPrefix .. i], "ANCHOR_LEFT", 0, -50);
				end
				
				local strQuestDescription, strQuestObjectives = GetQuestLogQuestText();
				local objColour = GetQuestDifficultyColor(level);
				GameTooltip:SetText(questTitle, 0, 1, 0, 1);
				GameTooltip:AddLine(strQuestObjectives, 0.5, 0.5, 0.5, 1);
				
				local strQuestTag = ""
				if (isDaily) then
					strQuestTag = strQuestTag .. L["Daily "]
				end				
				if (questTag) then
					strQuestTag = strQuestTag .. questTag
					if (suggestedGroup > 0) then
						strQuestTag = strQuestTag .. " (" .. suggestedGroup .. ")"
					end
				end				
				GameTooltip:AddLine(strQuestTag, 1, 0, 0, 1);
				
				local intPartyMembers = GetNumPartyMembers();
				local blnIsOnQuest = false
				local k = 1
				if (intPartyMembers > 0) then
					GameTooltip:AddLine(L["Party members on quest:"], 0, 1, 0, 1);
					for k = 1, intPartyMembers do
						blnIsOnQuest = IsUnitOnQuest(i, "party" .. k);
						if (blnIsOnQuest == 1) then
							GameTooltip:AddLine(UnitName("party" .. k), 1, 1, 1, 1);
						end
					end				
				end
				
				GameTooltip:Show();
			end
		end)
		dicButtons[strButtonPrefix .. i]:SetScript("OnLeave", function(self) 
			GameTooltip:Hide() 
		end)

	end
end

function SorhaQuestLog:UpdateMinion()
	local fraTempHeaderFrame = fraSQLMainAnchor
	local intHeaderPosition = 0
	local intQuestOffset = 10
	local intSpacingIncrease = 0
	local intLargestWidth = 0
	local blnLargestWidthIsHeader = false
	blnWasAClick = false
	
	local numEntries, numQuests = GetNumQuestLogEntries();	
	if (db.Frames.ShowNumberOfQuests == true) then
		fraSQLMainAnchor.objFontString:SetText(strHeaderColour .. numQuests .. "/25|r");
	else
		fraSQLMainAnchor.objFontString:SetText("");
	end
	fraSQLMainAnchor:SetWidth(db.Frames.MinionWidth)
	
	local i = 1
	for i=1,intNumberOfFrames,1 do
		dicButtons[strButtonPrefix .. i]:Hide()
		dicButtons[strButtonPrefix .. i].fraItemButton:Hide()
		dicButtons[strButtonPrefix .. i].objFontString:SetText("")
		dicButtons[strButtonPrefix .. i].objFontString:SetWidth(0)
		dicButtons[strButtonPrefix .. i].objFontString:SetHeight(0)
		dicButtons[strButtonPrefix .. i]:ClearAllPoints()
		dicButtons[strButtonPrefix .. i].isUsed = false
	end

	for i = 1, numEntries, 1 do
		local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(i);
		dicButtons[strButtonPrefix .. i].isHeader = isHeader
		dicButtons[strButtonPrefix .. i].isCollapsed = isCollapsed
		intSpacingIncrease = 0
		
		if (not isHeader) then
			if ((db.Frames.ShowAllQuests == true or (tContains(dbChar.Quests.HiddenQuests, questID) == nil)) or db.Frames.AllowHiddenQuests == false) then
				dicButtons[strButtonPrefix .. i]:Show()
				dicButtons[strButtonPrefix .. i].isUsed = true
				local strQuestText = self:GetQuestText(level, questTag, suggestedGroup, isDaily, questTitle, isComplete, i, questID)

				dicButtons[strButtonPrefix .. i]:SetPoint("TOP",fraSQLMainAnchor,"CENTER",intQuestOffset,intHeaderPosition)
				dicButtons[strButtonPrefix .. i].objFontString:SetFont("Interface\\AddOns\\SorhaQuestLog\\Fonts\\framd.ttf", db.Frames.QuestFontSize)
				dicButtons[strButtonPrefix .. i].objFontString:SetText(strQuestText);

				dicButtons[strButtonPrefix .. i]:SetWidth(db.Frames.MinionWidth - intQuestOffset*2)
				--
				if (dicButtons[strButtonPrefix .. i].objFontString:GetWidth()  > intLargestWidth) then
					intLargestWidth = dicButtons[strButtonPrefix .. i].objFontString:GetWidth()
					blnLargestWidthIsHeader = false
				end
				--
				dicButtons[strButtonPrefix .. i].objFontString:SetWidth(db.Frames.MinionWidth - intQuestOffset*2)
				intSpacingIncrease = dicButtons[strButtonPrefix .. i].objFontString:GetHeight()
			
				if (db.Frames.ShowItemButtons == true) then
					local link, item, charges = GetQuestLogSpecialItemInfo(i);
					if (item) then
						dicButtons[strButtonPrefix .. i].fraItemButton:ClearAllPoints();
						if (db.Frames.MoveItemsAndTooltipsRight == true) then
							dicButtons[strButtonPrefix .. i].fraItemButton:SetPoint("TOPLEFT", dicButtons[strButtonPrefix .. i], "TOPRIGHT", 7, 0)
						else
							dicButtons[strButtonPrefix .. i].fraItemButton:SetPoint("TOPRIGHT", dicButtons[strButtonPrefix .. i], "TOPLEFT", -12, 0)
						end
						dicButtons[strButtonPrefix .. i].fraItemButton:Show()
						dicButtons[strButtonPrefix .. i].fraItemButton:SetID(i)
						SetItemButtonTexture(dicButtons[strButtonPrefix .. i].fraItemButton, item);
						SetItemButtonCount(dicButtons[strButtonPrefix .. i].fraItemButton, charges);
						WatchFrameItem_UpdateCooldown(dicButtons[strButtonPrefix .. i].fraItemButton);
						
						-- If a button is heigher then its quest then expand the quest frame to stop overlapping buttons
						local intButtonHeight = dicButtons[strButtonPrefix .. i].fraItemButton:GetHeight() * dicButtons[strButtonPrefix .. i].fraItemButton:GetScale()
						if (intButtonHeight > intSpacingIncrease) then
							intSpacingIncrease = intButtonHeight
							dicButtons[strButtonPrefix .. i].objFontString:SetHeight(intButtonHeight)
						end
					end
				end
				dicButtons[strButtonPrefix .. i]:SetHeight(intSpacingIncrease)
			end
		else
			dicButtons[strButtonPrefix .. i]:Show()
			dicButtons[strButtonPrefix .. i].isUsed = true
			local strPrefix = strHeaderColour .. "- "
			if (isCollapsed == 1) then
				strPrefix = strHeaderColour .. "+ "
			end	

			dicButtons[strButtonPrefix .. i]:SetPoint("TOP",fraSQLMainAnchor,"CENTER",0,intHeaderPosition)
			dicButtons[strButtonPrefix .. i].objFontString:SetFont("Interface\\AddOns\\SorhaQuestLog\\Fonts\\framd.ttf", db.Frames.HeaderFontSize)
			dicButtons[strButtonPrefix .. i].objFontString:SetText(strPrefix .. questTitle .. "|r");
			
			dicButtons[strButtonPrefix .. i]:SetWidth(db.Frames.MinionWidth)
			--
			if (dicButtons[strButtonPrefix .. i].objFontString:GetWidth() > intLargestWidth) then
				intLargestWidth = dicButtons[strButtonPrefix .. i].objFontString:GetWidth()
				blnLargestWidthIsHeader = true
			end
			--
			dicButtons[strButtonPrefix .. i].objFontString:SetWidth(db.Frames.MinionWidth)
			
			intSpacingIncrease = dicButtons[strButtonPrefix .. i].objFontString:GetHeight()
			dicButtons[strButtonPrefix .. i]:SetHeight(intSpacingIncrease)
		end
		intHeaderPosition = intHeaderPosition - intSpacingIncrease 
    end
	
	if (db.Frames.CollapseToLeft == true) then
		local intNonQuestSize = 0
		local intQuestSize = 0
		if (blnLargestWidthIsHeader == true) then
			intNonQuestSize = intLargestWidth
			intQuestSize = intLargestWidth - (intQuestOffset * 2)
		else
			intNonQuestSize = intLargestWidth + (intQuestOffset * 2)
			intQuestSize = intLargestWidth	
		end
	
		if (intNonQuestSize < db.Frames.MinionWidth) then
			fraSQLMainAnchor:SetWidth(intNonQuestSize)
			for i = 1, numEntries, 1 do
				if (dicButtons[strButtonPrefix .. i].isUsed == true) then
					if (dicButtons[strButtonPrefix .. i].isHeader == 1) then
						dicButtons[strButtonPrefix .. i].objFontString:SetWidth(intNonQuestSize)
						dicButtons[strButtonPrefix .. i]:SetWidth(intNonQuestSize)
					else
						dicButtons[strButtonPrefix .. i].objFontString:SetWidth(intQuestSize)
						dicButtons[strButtonPrefix .. i]:SetWidth(intQuestSize)
					end
				end
			end
		end
	end
	blnIsUpdating = false
end

function SorhaQuestLog:GetQuestText(level, questTag, suggestedGroup, isDaily, questTitle, isComplete, i, questID)
	local objColour = GetQuestDifficultyColor(level);
	local strTitleColor = format("|c%02X%02X%02X%02X", 255, objColour.r * 255, objColour.g * 255, objColour.b * 255);
	local strReturnText = strTitleColor .. "[" .. level
	
	SelectQuestLogEntry(i);
	local numQuestLogLeaderBoards = GetNumQuestLeaderBoards(i)

	if (questTag == LFG_TYPE_DUNGEON) then
		strReturnText = strReturnText .. "d"
	end		
	if (questTag == RAID) then
		strReturnText = strReturnText .. "r"
	end		
	if (questTag == PVP) then
		strReturnText = strReturnText .. "p"
	end		
	if (suggestedGroup > 0) then
		strReturnText = strReturnText .. "g" .. suggestedGroup
	end
	if (isDaily) then
		strReturnText = strReturnText .. "*"
	end
	
	strReturnText = strReturnText .. "] " .. questTitle .. "|r"
	if (isComplete and isComplete < 0) then 
		strReturnText = strReturnText .. L[" (Failed)"]
	elseif (isComplete and isComplete > 0) then
		strReturnText = strReturnText .. L[" (Done)"]
	elseif (numQuestLogLeaderBoards == 0) then
		strReturnText = strReturnText .. L[" (goto)"]
	end
	
	if ((tContains(dbChar.Quests.HiddenQuests, questID) == 1) and db.Frames.AllowHiddenQuests == true) then
		strReturnText = strReturnText .. strHeaderColour .. L[" (Hidden)"] .. "|r"
	end 
	
	strReturnText = strReturnText .. "\n"

	local k = 1
	if (numQuestLogLeaderBoards > 0) then --If has objectives then for each objective print, else print quest 'objective' description
		for k=1,numQuestLogLeaderBoards,1 do
			local leaderboardTxt, itemType, isDone = GetQuestLogLeaderBoard(k,i);
			if (leaderboardTxt ~= nil) then --For some reason some quests have hidden objectives that are blank o_O
				if ((isDone == nil)) then
					strReturnText = strReturnText .. strObjectiveColour .. " - " .. leaderboardTxt .. "|r\n"
				else
					if (db.Frames.HideCompletedObjectives == false) then
						strReturnText = strReturnText .. strDoneObjectiveColour .. " - " .. leaderboardTxt .. "|r\n"
					end
				end	
			end
		end
	else 
		if (db.Frames.ShowDescWhenNoObjectives == true) then
			local strQuestDescription, strQuestObjectives = GetQuestLogQuestText();
			if (not(isComplete and isComplete > 0)) then
				strReturnText = strReturnText .. strObjectiveDescriptionColour .. " - " .. strQuestObjectives .. "|r\n"
			else
				if (db.Frames.HideCompletedObjectives == false) then
					strReturnText = strReturnText .. strObjectiveDescriptionColour .. " - " .. strQuestObjectives .. "|r\n"
				end
			end	
		end
	end
	strReturnText = strtrim(strReturnText)
	return strReturnText
end

function SorhaQuestLog:QUEST_LOG_UPDATE(...)
	if (db.Frames.HideSorhaQuestLog == false) then
		if (blnWasAClick == true) then -- Was the event called from a click (Skips the delay on header collapses etc)
			blnIsUpdating = true
			self:UpdateMinion()
		elseif (blnIsUpdating == false) then --If not updating then update, forces a 0.5 second delay between system called updates
			blnIsUpdating = true
			self:ScheduleTimer("UpdateMinion", 0.5)
		end
	end
end

function SorhaQuestLog:PLAYER_LEVEL_UP(...)
	if (db.Frames.HideSorhaQuestLog == false) then
		self:UpdateMinion()
	end
end

function SorhaQuestLog:SetupOptions() --Creates options. Registers chat commands
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SorhaQuestLog", getOptions())
	
	AceConfigDialog:SetDefaultSize("SorhaQuestLog", 680,525)
	self:RegisterChatCommand("sql", "ChatCommand")
	self:RegisterChatCommand("SorhaQuestLog", "ChatCommand")
end

function SorhaQuestLog:ChatCommand(input) --Handles chat commands
	if not input or input:trim() == "" then
		LibStub("AceConfigDialog-3.0"):Open("SorhaQuestLog")
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(SorhaQuestLog, "SorhaQuestLog", "SorhaQuestLog", input)
	end
end

function SorhaQuestLog:doColourChange() --Changes the addons colours on the fly as changed
	self:SetupColours()
	self:UpdateMinion()
end

function SorhaQuestLog:SetupColours()
	strHeaderColour = format("|c%02X%02X%02X%02X", 255, db.Frames.HeaderColourRed * 255, db.Frames.HeaderColourGreen * 255, db.Frames.HeaderColourBlue * 255);
	strObjectiveColour = format("|c%02X%02X%02X%02X", 255, db.Frames.ObjectiveColourRed * 255, db.Frames.ObjectiveColourGreen * 255, db.Frames.ObjectiveColourBlue * 255);
	strDoneObjectiveColour = format("|c%02X%02X%02X%02X", 255, db.Frames.ObjectiveDoneColourRed * 255, db.Frames.ObjectiveDoneColourGreen * 255, db.Frames.ObjectiveDoneColourBlue * 255);
	strObjectiveDescriptionColour = format("|c%02X%02X%02X%02X", 255, db.Frames.ObjectiveDescColourRed * 255, db.Frames.ObjectiveDescColourGreen * 255, db.Frames.ObjectiveDescColourBlue * 255);
end

function SorhaQuestLog:RemoveObject(objTable, objFind)
	for k, v in pairs(objTable) do
		if (v == objFind) then
			tremove(objTable,k)
			break
		end
	end
end

function SorhaQuestLog:doAnchorBackgroundUpdate()
	if db.Frames.MainWindowLocked == false then
		fraSQLMainAnchor:SetBackdropColor(0, 1, 0, 0.5)
		fraSQLMainAnchor:SetBackdropBorderColor(0.5, 0.5, 0, 0)
		fraSQLMainAnchor:EnableMouse(true)
	else
		fraSQLMainAnchor:SetBackdropColor(0, 1, 0, 0)
		fraSQLMainAnchor:SetBackdropBorderColor(0.5, 0.5, 0, 0)
		fraSQLMainAnchor:EnableMouse(false)
	end
end

function SorhaQuestLog:doHiddenQuestsUpdate()
	if (db.Frames.AllowHiddenQuests == true) then
		fraSQLMainAnchor.buttonShowHidden:Show()	
		fraSQLMainAnchor.buttonShowHiddenHover:Show()
		fraSQLMainAnchor.objFontString:ClearAllPoints()
		fraSQLMainAnchor.objFontString:SetPoint("TOPLEFT", fraSQLMainAnchor, "TOPLEFT", 16, -7);
		if (db.Frames.ShowAllQuests == true) then
			fraSQLMainAnchor.buttonShowHidden:SetBackdropColor(0, 0.3, 1, 1)
		else
			fraSQLMainAnchor.buttonShowHidden:SetBackdropColor(0.5, 0.5, 0.5, 1)
		end
	else
		fraSQLMainAnchor.buttonShowHidden:Hide()	
		fraSQLMainAnchor.buttonShowHiddenHover:Hide()
		fraSQLMainAnchor.objFontString:ClearAllPoints()
		fraSQLMainAnchor.objFontString:SetPoint("TOPLEFT", fraSQLMainAnchor, "TOPLEFT", 0, -7);
	end
	if (blnInitialized == true) then
		self:UpdateMinion()
	end
end

function SorhaQuestLog:doCreateLooseFrame(strType,strName,fraParent,intWidth,intHeight,intScale,strStrata,intLevel,intAlpha,finherit)
	local f = CreateFrame(strType,strName,fraParent,finherit)
	f:SetWidth(intWidth)
	f:SetHeight(intHeight)
	f:SetFrameStrata(strStrata)
	f:SetFrameLevel(intLevel)	
	f:SetAlpha(intAlpha)
	return f  
end 

function SorhaQuestLog:doCreateFrame(strType,strName,fraParent,intWidth,intHeight,intScale,strStrata,intLevel,strPoint,fraRelativeFrame,strRelativePoint,intOffsetX,intOffsetY,intAlpha,finherit)
	local f = CreateFrame(strType,strName,fraParent,finherit)
	
	f:SetWidth(intWidth)
	f:SetHeight(intHeight)
	f:SetFrameStrata(strStrata)
	f:SetFrameLevel(intLevel)
	f:SetPoint(strPoint,fraRelativeFrame,strRelativePoint,intOffsetX,intOffsetY)
	f:SetAlpha(intAlpha)
	return f  
end 

function SorhaQuestLog:ShowSorhaQuestLog()
	db.Frames.HideSorhaQuestLog = false
	fraSQLMainAnchor:Show()
	if (blnInitialized == true) then
		self:UpdateMinion()
	end
end

function SorhaQuestLog:HideSorhaQuestLog()
	db.Frames.HideSorhaQuestLog = true
	fraSQLMainAnchor:Hide()
end

function SorhaQuestLog:ToggleSorhaQuestLog()
	if (db.Frames.HideSorhaQuestLog == true) then
		self:ShowSorhaQuestLog()
	else
		self:HideSorhaQuestLog()
	end
end